#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<SDL.h>

#include "68000.h"
#include "st.h"
#include "mem.h"
#include "op68k.h"
#define min(A,B) (A)<(B)?(A):(B);
#define max(A,B) (A)>(B)?(A):(B);

#define tmaster     166234

//PSg variables
static int voltab[16]={0,0,0,0,1,1,2,2,4,6,8,13,19,28,43,64};

extern unsigned char sample[10000];

#define sampos psg[24]

int g_fillsound_len=0;
Uint8 *g_fillsound_stream;

void fillsoundbuffer ( void ) {
  int freq,vol,chan;
  int chans,envspd,env,epos;
  int buzfreq, nvol, nfreq;
  int freq2[3], vol2[3];
  register int wave;
  int fillsize= g_fillsound_len; /* 512 orig */
  int J, N=-1, I;
  unsigned char *nbuf = g_fillsound_stream;
	
  //Increment envelope counter for sound chip
  //PSG clock speed is 1789776hz
  //envelope repetition frequency=1789776/(256*freq)
  //so if envfreq was 139.82, that would be 50hz
  psg_epos+=140;
	
  //If sampos>40 then assume sampled sound is active
  if (sampos>40) {
    //Need to stretch 'sampos' samples to 'fillsize' buffer
    wave=(sampos<<16)/fillsize; sampos=0; I=fillsize*wave;
    //Fill buffer
    for (J=0; J<I; J+=wave) *nbuf++=sample[J>>16];
    //Done
    return;
  }
  sampos=0;

  //Get envelope, mixer register, and noise
  env=psg[13]&15;	chans=psg[7];
  envspd=(psg[12]<<8)+psg[11];
  buzfreq=0; nvol=0; 

  //Process all 3 channels
  for (chan=0; chan<3; chan++) {
    //Get volume
    vol=psg[8+chan]&31; 
    if (vol>=16) {
      //Buzzer effects
      if (envspd<256 && (env==8 || env==12)) {buzfreq=tmaster/(envspd*16+1); vol=0;}
      else if (envspd<128 && (env==10 || env==14)) {buzfreq=tmaster/(envspd*32+1); vol=0;}
      //Standard envelopes
      else {
	freq=((psg[chan+chan+1]&15)<<8)+psg[chan+chan];
	epos=psg_epos;
	if (envspd>0) {
	  if (env<=3 || env==9) {if (epos<envspd) vol=15-((15*epos)/envspd); else vol=0;}
	  else if (env<=7 || env==15) {if (epos<envspd) vol=(15*epos)/envspd; else vol=0;}
	  else if (env==11) {if (epos<envspd) vol=15-((15*epos)/envspd); else vol=15;}
	  else if (env==13) {if (epos<envspd) vol=(15*epos)/envspd; else vol=15;}
	  else if (env==8) vol=15-((15*(epos%envspd))/envspd);
	  else if (env==12) vol=(15*(epos%envspd))/envspd;
	  epos=epos%(envspd*2);
	  if (env==10) {if (epos<envspd) vol=15-((15*epos)/envspd); else vol=(15*(epos-envspd))/envspd;}
	  if (env==14) {if (epos<envspd) vol=(15*epos)/envspd; else vol=15-((15*(epos-envspd))/envspd);}
	}
	else {if (env<=3 || env==8 || env==9 || env==10 || env==11) vol=15; else vol=0;} 
	//Check mixer register
	if (chan==0) {
	  if ((chans&8)==0) nvol=vol;
	  if ((chans&1)==1) vol=0;
	}
	if (chan==1) {
	  if ((chans&16)==0 && vol>nvol) nvol=vol;
	  if ((chans&2)==2) vol=0;
	}
	if (chan==2) {
	  if ((chans&32)==0 && vol>nvol) nvol=vol;
	  if ((chans&4)==4) vol=0;
	}
      }
    }
    else {
      //Ordinary note
      freq=((psg[chan+chan+1]&15)<<8)+psg[chan+chan];
      //Check mixer register
      if (chan==0) {
	if ((chans&8)==0) nvol=vol;
	if ((chans&1)==1) vol=0;
      }
      if (chan==1) {
	if ((chans&16)==0 && vol>nvol) nvol=vol;
	if ((chans&2)==2) vol=0;
      }
      if (chan==2) {
	if ((chans&32)==0 && vol>nvol) nvol=vol;
	if ((chans&4)==4) vol=0;
      }
    }
    //Done! - save volume and frequency
    if (freq>10) {freq2[chan]=tmaster/(freq+1); vol2[chan]=voltab[vol];}
    else vol2[chan]=0;
  }

  //Prepare noise
  nvol=voltab[nvol]>>1;
  if (nvol>0) {nfreq=tmaster/((psg[6]&31)+1);}

  //Generate wave data
  for(I=0; I<fillsize; I++) {
//    wave=(I&16?0:255);
    wave=0; //128;
    //Normal channels

    phase0+=freq2[0]; if (phase0&16384) wave+=vol2[0]; 
    phase1+=freq2[1]; if (phase1&16384) wave+=vol2[1]; 
    phase2+=freq2[2]; if (phase2&16384) wave+=vol2[2];
    //'Mad-max' buzzer
    if (buzfreq>0) {phase3+=buzfreq; wave+=128+((phase3&16383)>>9);}
    //noise channel
    if (nvol>0) {
      if (nrand&1) wave+=nvol;
      phase4+=nfreq;
      if (phase4>32768) {
	if((nrand<<=1)&0x80000000) nrand^=0x08000001;
	phase4-=32768;
      }
    }

    //Output into buffer
    *nbuf++=wave & 0xFF;
  }

  //Send buffer to wave device

  return;
}

void audio_callback ( void *userdata, Uint8 *stream, int len ) {
  g_fillsound_stream = stream;
  g_fillsound_len = len;

  fillsoundbuffer();

  return;
}

static int audio_yetinit=0;

void audio_init(void)
{
	if (!audio_yetinit)
	{
		SDL_AudioSpec fmt;
		fmt.freq = 22050; //8000; //44100; //22050;
		fmt.format = AUDIO_U8;
		fmt.channels = 1;
		fmt.samples = 256; //1024; //512; // 128; //256; // 292;
		fmt.callback = audio_callback;
		fmt.userdata = NULL;

		if ( SDL_OpenAudio(&fmt, NULL) < 0 ) 
			fprintf(stderr, "Audio don't open!: %s\n",SDL_GetError());
		else
			audio_yetinit=1;
	}
	if (audio_yetinit)
	{
		lastpsg=10;
		SDL_PauseAudio(0);
	}
}

void audio_stop(void)
{
	SDL_PauseAudio(1);
}
