/*
* Castaway
*  (C) 1994 - 2002 Martin Doering, Joachim Hoenig
*
* $File$ - memory read/write
*
* This file is distributed under the GPL, version 2 or at your
* option any later version.  See doc/license.txt for details.
*
* revision history
*  23.05.2002  JH  FAST1.0.1 code import: KR -> ANSI, restructuring
*  30.05.2002  JH  Discontinued using mmap and mprotect, now using
*                  Martin's memory access jump table.
*  12.06.2002  JH  Correct bus error/address error exception stack frame
*  14.06.2002  JH  LowRamSetX() functions improved.
*  09.07.2002  JH  Now loads any 192k ROM file
*  10.07.2002  MAD Now loads any ROM file
*  16.09.2002  JH  Bugfix: Word access on unmapped I/O address stacked
*                  two bus error stack frames. Fault address corrected.
*  08.10.2002  JH  Fixed integer types.
*  27.10.2002  AG  Trashed everything for more speed! mwuhahaha!
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#ifndef DREAMCAST
#include <memory.h>
#else
#include <string.h>
#endif
#include "config.h"
#include "68000.h"
#include "st.h"
#include "mem.h"

int8 *membase=0;
int8 *rombase=0;
uint16 tosnumber=0;

char rom[80] = ROM;
static int samvol[16]={0,0,0,1,1,1,2,3,5,7,11,17,25,38,57,85};
extern uint32 psg[26];
#define lastpsg psg[25]
#define sampos psg[24]
char GetMemB(unsigned long address)
{
	address &= MEMADDRMASK;
	if (address<MEMSIZE) {
		//RAM
		return ReadB(address + membase);
	}
	if (address>=ROMBASE && address<ROMBASE+ROMSIZE) {
		//ROM
		return ReadB(address + rombase);
	}
	if (address>=IOBASE && address<IOBASE+IOSIZE) {
		//IO
		return DoIORB(address);
	}
	//Unmapped
	ON_UNMAPPED(address, value);
	return -1;
}

/* Fetch word, address may not be word-aligned */
short GetMemW(unsigned long address)
{
#ifdef CHKADDRESSERR
    address &= MEMADDRMASK;
    if (address & 0x1){
		ExceptionGroup0(ADDRESSERR, address, 1);
		return -1;
	}
#else
    address &= MEMADDRMASKS;
#endif
	if (address<MEMSIZE) {
		//RAM
		return ReadW(address + membase);
	}
	if (address>=ROMBASE && address<ROMBASE+ROMSIZE) {
		//ROM
		return ReadW(address + rombase);
	}
	if (address>=IOBASE && address<IOBASE+IOSIZE) {
		//IO
		return DoIORW(address);
	}
	//Unmapped
	ON_UNMAPPED(address, value);
	return -1;
}

/* Fetch dword, address may not be dword-aligned */
long GetMemL(unsigned long address)
{
#ifdef CHKADDRESSERR
    address &= MEMADDRMASK;
    if (address & 0x1){
		ExceptionGroup0(ADDRESSERR, address, 1);
		return -1;
	}
#else
    address &= MEMADDRMASKS;
#endif
	if (address<MEMSIZE) {
		//RAM
		return ReadL(address + membase);
	}
	if (address>=ROMBASE && address<ROMBASE+ROMSIZE) {
		//ROM
		return ReadL(address + rombase);
	}
	if (address>=IOBASE && address<IOBASE+IOSIZE) {
		//IO
		return DoIORL(address);
	}
	//Unmapped
	ON_UNMAPPED(address, value);
	return -1;
}

/* Write byte to address */
void SetMemB (unsigned long address, unsigned char value) {
    address &= MEMADDRMASK;
	ON_WRITE(address, value);
	if (address<MEMSIZE) {
		//RAM
		if (address<SVADDR && !GetFC2()){
			ExceptionGroup0(BUSERR, address, 0);
			return;
		}
		WriteB(address + membase, value);
		return;
	}
	if (address>=IOBASE && address<IOBASE+IOSIZE) {
		//IO
		DoIOWB(address, value);
		return;
	}
	//Unmapped
	ON_UNMAPPED(address, value);
}

/* Write word, address may not be word-aligned */
void SetMemW(unsigned long address, unsigned short value)
{	
#ifdef CHKADDRESSERR
    address &= MEMADDRMASK;
    if (address & 0x1){
		ExceptionGroup0(ADDRESSERR, address, 1);
		return;
	}
#else
    address &= MEMADDRMASKS;
#endif
	ON_WRITE(address, value);
	if (address<MEMSIZE) {
		//RAM
		if (address<SVADDR ){
			if (!GetFC2()||address<8){
				ExceptionGroup0(BUSERR, address, 0);
				return;
			}
		}
		
		WriteW(address + membase, value);
		return;
	}
	if (address>=IOBASE && address<IOBASE+IOSIZE) {
		//IO
		DoIOWW(address, value);
		return;
	}
	//Unmapped
	ON_UNMAPPED(address, value);
}

/* Write dword, address may not be dword-aligned */
void SetMemL(unsigned long address, unsigned long value)
{
#ifdef CHKADDRESSERR
    address &= MEMADDRMASK;
    if (address & 0x1){
		ExceptionGroup0(ADDRESSERR, address, 1);
		return;
	}
#else
    address &= MEMADDRMASKS;
#endif	
	
	ON_WRITE(address, value);
	if (address<MEMSIZE) {
		//RAM
		if (address<SVADDR){
			if (!GetFC2()||address<8){
				ExceptionGroup0(BUSERR, address, 0);
				return;
			}
		}
		
		WriteL(address + membase, value);
		return;
	}
	if (address>=IOBASE && address<IOBASE+IOSIZE) {
		//IO
		DoIOWL(address, value);
		return;
	}
	//Unmapped
	ON_UNMAPPED(address, value);
}

//Simplifed versions of memory access commands used for instructions fetch
//Instruction fetch is likely only to be from ram or rom, otherwise the
//program will surely have crashed anyway!

char GetMemBpc(unsigned long address)
{
	address &= MEMADDRMASK;
	if (address<MEMSIZE) return ReadB(address + membase);
	else return ReadB(address + rombase);
}

short GetMemWpc(unsigned long address)
{
	address &= MEMADDRMASK;
	if (address<MEMSIZE) return ReadW(address + membase);
	else return ReadW(address + rombase);
}

/* Fetch dword, address may not be dword-aligned */
long GetMemLpc(unsigned long address)
{
	address &= MEMADDRMASK;
	if (address<MEMSIZE) return ReadL(address + membase);
	else return ReadL(address + rombase);
}

static void *__rombase=NULL, *__membase=NULL;

void            MemQuit(void)
{
	if (__membase)
		free(__membase);
	__membase=NULL;
	if (__rombase)
		free(__rombase);
	__rombase=NULL;
}

void MemClean(void)
{
	int n;
	membase=__membase;

	for (n=0; n<MEMSIZE+4; n++) membase[n]=0;
	switch (tosnumber){
	case 0x0001:
	case 0x0201:
	case 0x0401:
		/* Copy first 8 bytes of ROM to RAM */
		memcpy (membase, rombase+ROMBASE, 8);   
		break;
	case 0x0502:
	case 0x0602:
		/* Copy first 8 bytes of ROM to RAM */
		memcpy (membase, rombase+ROMBASE2, 8);   
		break;
	}
}



int MemInit(void)
{
	int n;
	uint8 val1,val2;
	FILE *roms;

	MemQuit();
    //Load ROM
	if (NULL == (roms = fopen (rom, "rb"))) {
		return 1;
	}
	fread (&tosnumber, 1, 2, roms);
	fread (&tosnumber, 1, 2, roms);
	(void) fclose (roms);

	roms = fopen (rom, "rb");

	//printf( "TOS NUM: %x %x", *((char*)&tosnumber), *(((char*)&tosnumber)+1)  );
	//do_wait_for_key();
	
	switch (tosnumber){
	case 0x0001:
	case 0x0201:
	case 0x0401:
		rombase = malloc(ROMSIZE+4);
		__rombase=rombase;
		if (ROMSIZE != fread (rombase, 1, ROMSIZE, roms)) {
			return 2;
		}
		break;
	case 0x0502:
	case 0x0602:
		rombase = malloc(ROMSIZE2+4);
		__rombase=rombase;
		if (ROMSIZE2 != fread (rombase, 1, ROMSIZE2, roms)) {
			free(__rombase);
			__rombase=NULL;
			return 3;
		}
		break;
	}
	(void) fclose (roms);
#ifdef BYTES_SWAP
	switch (tosnumber){
	case 0x0001:
	case 0x0201:
	case 0x0401:
		for (n=0; n<ROMSIZE; n+=2) {
			val1 = rombase[n];
			val2 = rombase[n+1];
			rombase[n] = val2;
			rombase[n+1] =val1;
		}
		break;
	case 0x0502:
	case 0x0602:
		for (n=0; n<ROMSIZE2; n+=2) {
			val1 = rombase[n];
			val2 = rombase[n+1];
			rombase[n] = val2;
			rombase[n+1] =val1;
		}
		break;
	}
#endif
	/* precalculate rombase - now just address-adding happens later */
	switch (tosnumber){
	case 0x0001:
	case 0x0201:
	case 0x0401:
		rombase -= ROMBASE;
		break;
	case 0x0502:
	case 0x0602:
		rombase -= ROMBASE2;
		break;
	}
	
	//Allocate and clear RAM
	membase = malloc(MEMSIZE+4);
	__membase=membase;
	MemClean();
	return 0;
}

//Movep support
//-------------

void SetMemPW(unsigned long address, unsigned long value)
{	int8 psgctrl1,value1;

if (address&0xffff03==0xff8800) {
	uint32 psgctrl1=(value>>8)&15;
	uint32 value1=value&255;
	psg[psgctrl1]=value1;
	if (psgctrl1==lastpsg) sample[sampos++]=samvol[psg[8]]+samvol[psg[9]]+samvol[value1];
	else if (psgctrl1==13 || psgctrl1==12) psg_epos=0;
	return;
}
address &= MEMADDRMASK;
if (address>=IOBASE && address<IOBASE+IOSIZE) {
	//IO
	DoIOWB(address, (int8)(value>>8));
	DoIOWB(address+2, (int8)value);
	return;
}
if (address<MEMSIZE) {
	//RAM
	if (address<SVADDR && !GetFC2()) ExceptionGroup0(BUSERR, address, 0);
	address+=(uint32)membase;
	WriteB(address, (int8)(value>>8));
	WriteB(address+2, (int8)value);
	return;
}
}

void SetMemPL(unsigned long address, unsigned long value)
{	int8 psgctrl1,psgctrl2,value1,value2;

if (address&0xffff03==0xff8800) {
	uint32 psgctrl1=(value>>24)&15;
	uint32 value1=(value>>16)&255;
	psg[psgctrl1]=value1;
	if (psgctrl1==lastpsg) sample[sampos++]=samvol[psg[8]]+samvol[psg[9]]+samvol[value1];
	else if (psgctrl1==13 || psgctrl1==12) psg_epos=0;
	psgctrl1=(value>>8)&15;
	value1=value&255;
	if (psgctrl1==lastpsg) sample[sampos++]=samvol[psg[8]]+samvol[psg[9]]+samvol[value1];
	else if (psgctrl1==13 || psgctrl1==12) psg_epos=0;
	return;
}
address &= MEMADDRMASK;
if (address>=IOBASE && address<IOBASE+IOSIZE) {
	//IO
	DoIOWB(address, (int8)(value>>24));
	DoIOWB(address+2, (int8)(value>>16));
	DoIOWB(address+4, (int8)(value>>8));
	DoIOWB(address+6, (int8)value);
	return;
}
if (address<MEMSIZE) {
	//RAM
	if (address<SVADDR && !GetFC2()) ExceptionGroup0(BUSERR, address, 0);
	address+=(uint32)membase;
	WriteB(address, (int8)(value>>24));
	WriteB(address+2, (int8)(value>>16));
	WriteB(address+4, (int8)(value>>8));
	WriteB(address+6, (int8)value);
	return;
}
}


